;; Generated package description from ada-mode.el
(define-package "ada-mode" "5.3.2" "major-mode for editing Ada sources" '((wisi "1.1.6") (cl-lib "0.4") (emacs "24.3")) :url "http://www.nongnu.org/ada-mode/" :keywords '("languages" "ada"))
