;;; ada-xref.el --- Dummy file to hide obsolete bundled version  -*- lexical-binding:t -*-
;;; see xref-ada.el for Emacs 25 xref minor mode for ada-mode
(require 'ada-mode)
(provide 'ada-xref)
(message "'ada-xref' is obsolete; use 'ada-mode' instead")
